// Copyright (C) 2004, Brian Enigma <enigma at netninja.com>
// This file is part of iGallery.
//
// iGallery is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// iGallery is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Foobar; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package org.ninjasoft.igallery.logic;

import java.util.*;
import java.net.*;
import java.io.*;

import org.ninjasoft.igallery.exceptions.*;
import org.ninjasoft.igallery.data.*;

/**
 * Class to wrap a gallery connection
 * @author enigma
 */
public class GalleryConnection {
    private StatusListener listener = null;
    private Cookiejar cookiejar = new Cookiejar();
    private URL url;
    public static boolean DEBUG = false;
    
    public GalleryConnection(String url) throws MalformedURLException {
        this(url, null);
    }
    
    public GalleryConnection(String url, StatusListener listener) throws MalformedURLException {
        this.listener = listener;
        this.url = new URL(url);
        if (!this.url.getProtocol().equals("http"))
            throw new MalformedURLException("This application only supports standard http:// connections");
    }
    
    /**
     * Login with the given username and password.  The login information is then cached
     * for the remainder of this object's usage.  If successful, return true.  If fail,
     * throw exception.  (This function actually never returns a false, only throws an
     * exception.)
     * @param username
     * @param password
     * @return
     * @throws IOException
     * @throws GalleryException
     */
    public boolean login(String username, String password) throws IOException, GalleryException {
        Properties parameters = new Properties();
        parameters.put("cmd", "login");
        parameters.put("protocol_version", "2.0");
        parameters.put("uname", username);
        parameters.put("password", password);
        Properties returnData = runCommand(parameters);
        checkResponseCode(returnData);
        
        return true;
    }
    
    /**
     * Return a list of albums available to the user.  
     * @return Hashtable with name=String of album name, value=Album object
     * @throws IOException
     * @throws GalleryException
     */
    public Hashtable getAlbumList() throws IOException, GalleryException {
        Properties parameters = new Properties();
        parameters.put("cmd", "fetch-albums-prune");
        parameters.put("protocol_version", "2.2");
        parameters.put("check_writable", "no");
        Properties returnData = runCommand(parameters);
        checkResponseCode(returnData);
        return parseAlbumList(returnData);
    }
    
    /**
     * Given the server response for fetch-albums-prune, return a set of
     * album beans.  This is returned in a Hashtable with the name
     * as the key and the bean as the object (this is so that the parent
     * album of nested albums can be determined because the parent
     * field is returned as a refnumber)
     * @param data
     * @return
     */
    private Hashtable parseAlbumList(Properties data) {
        Hashtable result = new Hashtable();
        int albumCount = 0;
        try{
        	albumCount = Integer.parseInt(data.getProperty("album_count", "0"));
        }catch(Exception e){}
        // Create the list
        for (int i=1; i<=albumCount; i++) {
            GalleryAlbum album = new GalleryAlbum();
            album.setRefNumber(i);
            album.setName(data.getProperty("album.name." + i, ""));
            album.setTitle(data.getProperty("album.title." + i, ""));
            album.setSummary(data.getProperty("album.summary." + i, ""));
            album.setParent(data.getProperty("album.parent." + i, ""));
            result.put(album.getName(), album);
        }
        // Interleave the items in the list, such that we can build a sort of tree or breadcrumb trail
        for (Iterator i = result.values().iterator(); i.hasNext(); ) {
            GalleryAlbum thisAlbum = (GalleryAlbum) i.next();
            String parent = "";
            try{
            	parent = thisAlbum.getParent();
            }catch(Exception e){}
            if (parent != null)
                thisAlbum.setParentAlbum((GalleryAlbum) result.get(parent));
        }
        return result;
    }
    
    public void uploadPhoto(GalleryAlbum album, File image, String name, String caption) throws GalleryException, IOException {
        Properties parameters = new Properties();
        parameters.put("cmd", "add-item");
        parameters.put("protocol_version", "2.0");
        parameters.put("set_albumName", album.getName());
        parameters.put("userfile_name", name);
        parameters.put("userfile", fileContent(image));
        parameters.put("force_filename", name);
        if ((caption != null) && (caption.length() > 0))
            parameters.put("caption", caption);
        Properties returnData = runCommand(parameters);
        checkResponseCode(returnData);
    }
    
    /**
     * Given a file, return the encoded file contents. 
     * @param f
     * @return
     * @throws IOException
     */
    public String encodeFileContent(File f) throws IOException {
        StringBuffer result = new StringBuffer();
        FileInputStream in = new FileInputStream(f);
        int c;
        while ((c = in.read()) != -1) {
            char ch = (char) c;
            if ( ((ch >= '0') && (ch <= '9')) ||
                 ((ch >= 'a') && (ch <= 'z')) ||
                 ((ch >= 'A') && (ch <= 'Z'))
               ) {
            	result.append(ch);
            } else {
                result.append("%");
                String code = Integer.toString(c, 16);
                if (code.length() == 1)
                    result.append("0");
                result.append(code);
            }
        }
        in.close();
        return result.toString();
    }
    
    /**
     * Given a file, return the file contents. 
     * @param f
     * @return
     * @throws IOException
     */
    public byte[] fileContent(File f) throws IOException {
        ByteArrayOutputStream result = new ByteArrayOutputStream();
        FileInputStream in = new FileInputStream(f);
        byte[] buff = new byte[1024];
        int len;
        while ((len = in.read(buff)) > 0)
            result.write(buff, 0, len);
        in.close();
        result.close();
        return result.toByteArray();
    }

    /**
     * Given an album, return a list of pictures in that album.
     * @param album
     * @return Hashtable with name=String photo name, value=Image object
     * @throws IOException
     * @throws GalleryException
     */
    public Hashtable getPhotoList(GalleryAlbum album) throws IOException, GalleryException {
        Properties parameters = new Properties();
        parameters.put("cmd", "fetch-album-images");
        parameters.put("protocol_version", "2.4");
        parameters.put("set_albumName", album.getName());
        parameters.put("albums_too", "no");
        Properties returnData = runCommand(parameters);
        checkResponseCode(returnData);
        return parseImageList(returnData);
    }
    
    /**
     * Given the server response for fetch-album-images, return a set of
     * image beans.  This is returned in a Hashtable with the name
     * as the key and the bean as the object.
     * @param data
     * @return
     */
    private Hashtable parseImageList(Properties data) {
        Hashtable result = new Hashtable();
        int albumCount = 0;
        try{
            albumCount = Integer.parseInt(data.getProperty("image_count", "0"));
        }catch(Exception e){}
        // Create the list
        for (int i=1; i<=albumCount; i++) {
            GalleryImage image= new GalleryImage();
            image.setName(data.getProperty("image.name." + i, ""));
            image.setCaption(data.getProperty("image.caption." + i, ""));
            result.put(image.getName(), image);
        }
        return result;
    }

    /**
     * Ensure that the response code is present and equal to zero.  Throw the proper
     * exception if it is an error code.
     * @param data
     * @throws GalleryException
     */
    private void checkResponseCode(Properties data) throws GalleryException {
        String codeStr = data.getProperty("status", "-1");
        String message = data.getProperty("status_text", "");
        int code = -1;
        try{
        	code = Integer.parseInt(codeStr);
        }catch(Exception e){}
        if (DEBUG)
        	sendStatus("Got from server: " + code + " (" + message + ")");
        if (code == -1)
            throw new GalleryException(-1, "Could not find status in data");
        if (code == 0)
            return;
        switch(code) {
            case 101:
            case 102:
            case 103:
            case 104:
                throw new InvalidVersionException(code, message);
            case 201:
            case 202:
                throw new InvalidLoginException(code, message);
            case 301:
                throw new UnknownCommandException(code, message);
            case 401:
            case 404:
            case 501:
                throw new InvalidPermissionException(code, message);
            case 402:
                throw new NoFilenameException(code, message);
            case 403:
                throw new UploadFailException(code, message);
            case 502:
                throw new CreateFailedException(code, message);
            default:
                throw new GalleryException(code, message);
        }
    }
    
    /**
     * Given a bunch of name/value pairs, send them to the server and get a set of
     * name value pairs back.
     * @param parameters
     * @return
     * @throws IOException
     * @throws GalleryException
     */
    private Properties runCommand(Properties parameters) throws IOException, GalleryException {
        byte[] queryString = null;
        HttpURLConnection connection = (HttpURLConnection) this.url.openConnection();
        if (!parameters.getProperty("cmd", "").equals("add-item")) {
            queryString = buildQueryString(parameters).getBytes();
        } else {
            String boundary = "----iGallery-multipart-boundary----";
            connection.addRequestProperty("Content-type", "multipart/form-data;.boundary=" + boundary);
            queryString = buildMultipart(parameters, boundary);
        }
        int len = queryString.length;
        connection.setDoOutput(true);
        connection.setDoInput(true);
        cookiejar.setCookies(connection);
        connection.addRequestProperty("Content-length", Integer.toString(len));
        OutputStream os = connection.getOutputStream();
        os.write(queryString, 0, len);
        os.close();
        connection.connect();
        if (connection.getResponseCode() != 200) {
            throw new WebServerException(-1, "Invalid response from webserver: expected 200, got " + 
                    connection.getResponseCode() + " (" +
                    connection.getResponseMessage() + ")");
        }
        InputStream in = connection.getInputStream();
        Properties results = parseResults(in);
        cookiejar.grabCookies(connection);
        return results;
    }
    
    /**
     * Given a server response, parse it into a Properties object
     * @param in
     * @return
     * @throws IOException
     */
    private Properties parseResults(InputStream in) throws IOException {
        Properties results = new Properties();
        // If debugging, dump the message to stdout
        if (DEBUG) {
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            byte buff[] = new byte[1024];
            int len;
            while ((len = in.read(buff)) > 0)
                out.write(buff, 0, len);
            out.close();
            if (DEBUG) {
            	sendStatus("Got the following data from the server:");
            	sendStatus(out.toString());
            }
            in = new ByteArrayInputStream(out.toByteArray());
        }
        // TODO: Skip bytes until start "#__GR2PROTO__
        results.load(in);
        return results;
    }
    
    /**
     * Given a Properties object of data to send to the server, turn it into
     * POST data.
     * @param properties
     * @return
     */
    private String buildQueryString(Properties properties) {
        StringBuffer result = new StringBuffer();
        for (Iterator i = properties.keySet().iterator(); i.hasNext(); ) {
            String key = (String) i.next();
            String value = properties.getProperty(key, "");
            result.append(URLEncoder.encode(key));
            result.append("=");
            result.append(URLEncoder.encode(value));
            result.append("&");
        }
        return result.toString();
    }
    
    private byte[] buildMultipart(Properties properties, String boundary) throws IOException {
        ByteArrayOutputStream result = new ByteArrayOutputStream();
        for (Iterator i = properties.keySet().iterator(); i.hasNext(); ) {
            String key = (String) i.next();
            Object value = properties.get(key);
            result.write("--".getBytes());
            result.write(boundary.getBytes());
            result.write("\r\n".getBytes());
            result.write("Content-Disposition: form-data; name=\"".getBytes());
            result.write(key.getBytes());
            result.write("\"".getBytes());
            if (value instanceof byte[]) {
                String filename = properties.getProperty("force_filename", "");
                result.write("; filename=\"".getBytes());
                result.write(filename.getBytes());
                result.write("\"".getBytes());
                /*
                if (filename.toLowerCase().endsWith(".jpg"))
                	result.write("\r\nContent-Type: image/jpeg".getBytes());
                else if (filename.toLowerCase().endsWith(".gif"))
                    result.write("\r\nContent-Type: image/gif".getBytes());
                else if (filename.toLowerCase().endsWith(".png"))
                    result.write("\r\nContent-Type: image/png".getBytes());
                */
            }
            result.write("\r\n\r\n".getBytes());
            if (value instanceof String)
            	result.write(((String) value).getBytes());
            else if (value instanceof byte[])
                result.write((byte[]) value);
            result.write("\r\n".getBytes());
        }
        result.write("--".getBytes());
        result.write(boundary.getBytes());
        result.write("--\r\n".getBytes());
        return result.toByteArray();
    }
    
    /**
     * Simple unit test
     * @param argv
     * @throws Exception
     */
    public static void main(String[] argv) throws Exception {
        GalleryConnection con = new GalleryConnection("http://localhost/gallerytest/gallery_remote2.php");
        con.login("admin", "test");
        Hashtable albums = con.getAlbumList();
        for (Iterator i = albums.values().iterator(); i.hasNext(); ) {
            GalleryAlbum thisAlbum = (GalleryAlbum) i.next();
            System.out.println(thisAlbum);
        }
        System.out.println("Uploading first photo...");
        con.uploadPhoto((GalleryAlbum) albums.get("nest2"), new File("/tmp/pics/IMG_1811.JPG"), "Pic1.jpg", "Caption 1");
        System.out.println("Uploading second photo...");
        con.uploadPhoto((GalleryAlbum) albums.get("nest2"), new File("/tmp/pics/IMG_1818.JPG"), "Pic2.jpg", "Caption 2");
    }

    public void setStatusListener(StatusListener l)     {this.listener = l;}

    private void sendStatus(String message) {
        if (listener != null)
            listener.setStatus(message);
    }
    private void sendError(String message) {
        if (listener != null)
            listener.setError(message);
    }
}
